unit ICMiscAppleEvents;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the support for a couple of miscellaneous AppleEvents
	supported by IC.  The first is the ICEditPreferences AppleEvent. This AppleEvent
	is sent by the Internet Config Extension in response to an ICEditPreferences API call.
	The purpose of the event is to instruct the application to show a specific preference
	to the user.  A good example of this is MacWeb 2.0, which doesn't provide an
	interface for editing Internet Config preference. Instead it shows the preference
	name in blue underlined text. When you click on the text, it calls ICEditPreference,
	which launches Internet Config and opens the window associated with that preference.
	
	The second AppleEvent we handle here is the GURL event, the Macintosh
	standard for asking an application to fetch a URL.  IC handles this AppleEvent and
	acts as a router, immediately forwarding the AppleEvent along to the helper
	application that's configured in the open IC document (or in the default 
	Internet Preferences if there is no open document).
	
	This module has only one entry point, an initialisation routine that is called to setup
	the AppleEvent handlers. The rest of the code is internal.
*)

interface

	procedure InstallMiscAppleEventHandlers;
		(* This routine installs the AppleEvent handler in this module. *)

implementation

	uses
		AppleEvents,
		TextUtils,
		Resources,
		
		ICTextWhats,
		ICHelperWhat,
		ICFileMapWhat,
		
		ICResources,
		ICMiscSubs,
		ICCommonSubs,
		ICGlobals, 
		ICDocument, 
		ICWindowGlobals, 
		ICWindows;

	(* **** Description of the AppleEvents ***** *)
	
	const
		kICEditPreferenceEventClass = 'ICAp';
		
		kICEditPreferenceEvent = 'ICAp';
			(* keyDirectObject *)
				(* This is the key string we're asked to open. *)
			keyEditPreferenceDestination = 'dest';
				(* This is the file you want to edit. Leave it empty to get the currently
					open preferences file, or the default preferences file (ie Internet 
					Preferences in the Preferences folder) if there's none open.
					Pass true to get the default preferences file.  Pass the FSSpec
					if you know which file you want to edit.
				*)
				
		kGURLEventClass = 'GURL';
		
		kGURLEvent = 'GURL';
			(* keyDirectObject *)
				(* This is the URL we're asked to open. *)
			(* We don't handle any of the other (optional) parameters to GURL
				because there's no easy way we can pass them along to the final
				application (ICLaunchURL takes a URL and that's it).
			*)

	(* **** Functional Implementation ***** *)

	function IsPrefix(var candidateKey : Str255; key : Str255) : Boolean;
		(* Returns true if key is a prefix to candidateKey. If we return true,
			also remove key from the front of candidate key.
			
			For example, IsPrefix('HelperFrog', 'Helper') returns true 
			and sets candidateKey to 'Frog'.
		*)
	begin
		IsPrefix := false;
		if EqualString(TPCopy(candidateKey, 1, length(key)), key, false, true) then begin
			candidateKey := TPCopy(candidateKey, length(key) + 1, 255);
			IsPrefix := true;
		end; (* if *)
	end; (* IsPrefix *)
	
	function FindWindowAndSelectKey (key: Str255): OSStatus;
		(* This routine is the 'sharp end' of the ICEditPreference API.
			It finds and opens the window that contains the What that edits this key.  Also,
			if the What is a TEXT What, select the text. If the key is one of the two
			structured keys, pass the key off to the structured keys handler to
			select the remainder of the key.
		*)
		var
			err: OSStatus;
			wt: WindowType;
			anItem : integer;
			whatTemplateID : integer;
			whatTemplate : WhatTemplateArrayHandle;
			s : SInt8;
	begin
		err := -1;		(* Default error value. *)
		
		(* First take out the special cases. *)
		if IsPrefix(key, concat(kICMapping,'')) then begin
			err := WindowOpen(WT_FileMapping);
			if err = noErr then begin
				MappingSetSelection(key);
			end; (* if *)
		end else if IsPrefix(key, kICHelper) then begin
			err := WindowOpen(WT_Helper);
			if err = noErr then begin
				HelperSetSelection(key);
			end; (* if *)
		end else begin

			(* Now search through the windows looking for a key that matches. 
				Note that we search the What template resources, not the Whats
				in memory. This is because the Whats in memory are built 'on
				the fly', when you open windows. So, if there are no windows open,
				there is no useful information in gWindows[wt].items. Instead
				we get the template resources and search them for the key.
			*)
			for wt := WT_Personal to pred(WT_Last) do begin
				whatTemplateID := MapWindowTypeToResourceID(wt);
				whatTemplate := WhatTemplateArrayHandle(Get1Resource('WHAT', whatTemplateID));
				if whatTemplate <> nil then begin
					s := HGetState(Handle(whatTemplate));
					HLock(Handle(whatTemplate));
					for anItem := 1 to GetHandleSize(Handle(whatTemplate)) div sizeof(WhatTemplateRecord) do begin
						if EqualString(key, whatTemplate^^[anItem].templateKey, false, true) then begin
							err := WindowOpen(wt);
							if (err = noErr) & (whatTemplate^^[anItem].templateTypeID = 'TEXT') then begin
								SelectTextItem(wt, anItem);
							end; (* if *)
						end; (* if *)
					end; (* for *)
					HSetState(Handle(whatTemplate), s);
				end; (* if *)
			end; (* for *)
			
		end; (* if *)
		
		FindWindowAndSelectKey := err;
	end; (* FindWindowAndSelectKey *)

	function DoEditPreference (key: Str255; fileSpec: FSSpec; 
													useFileSpec : Boolean; useCurrentFile : Boolean): OSStatus;
		(* Implements the edit preference AppleEvent that's the back end to the ICEditPreference
			API routine. The purpose of this routine is to switch to the appropriate preferences
			file and select the user interface that edits the key preference. key can be
			empty to just open the file and display the main window.

			If useFileSpec is true, this routine switches to the preferences file denoted by fileSpec.
			Otherwise it attempts to use the file associated with the current open document.
			If there isn't a current open document, it uses the default Internet Preferences
			document.
		*)
		var
			err: OSStatus;
			currentDocumentFile : FSSpec;
	begin
		(* Normalise the input parameters, ie setup fileSpec based on useFileSpec and
			useCurrentFile. Note that GetDocumentCurrentFile only returns valid information
			if IsDocumentOpen returns true. It's not immediately obvious, but this routine
			only relies on currentDocumentFile inside code that's guarded by IsDocumentOpen
			calls.
		*)
		GetDocumentCurrentFile(currentDocumentFile);
		err := noErr;
		if not useFileSpec then begin
			useCurrentFile := useCurrentFile & IsDocumentOpen;		(* Can only useCurrentFile if a document is open. *)
			if useCurrentFile then begin
				fileSpec := currentDocumentFile;
			end else begin
				err := FindInternetPreferences(fileSpec);
			end; (* if*)
		end; (* if *)
		
		(* Now switch to the correct preferences file. *)
		if err = noErr then begin
			if not IsDocumentOpen
					 | (fileSpec.vRefNum <> currentDocumentFile.vRefNum)
					 | (fileSpec.parID <> currentDocumentFile.parID) 
					 | not EqualString(fileSpec.name, currentDocumentFile.name, false, true) then begin
				err := DoOpenDocument(fileSpec);
			end; (* if *)
		end; (* if *)
		
		(* If all is OK, select the appropriate preference. *)
		if err = noErr then begin
			if key <> '' then begin
				err := FindWindowAndSelectKey(key);
			end; (* if *)
		end; (* if *)

		DoEditPreference := err;
	end; (* DoEditPreference *)

	(* **** AppleEvent Handler ***** *)

{$PUSH}
{$R-}
	function HandleAEEditPreferenceEvent(var appleEvent : AppleEvent; var reply: AppleEvent; refcon: longint): OSErr;
		var
			err: OSStatus;
			junkType: DescType;
			junkActualSize: Size;
			actualSize : Size;
			fileSpec: FSSpec;
			useFileSpec : Boolean;
			useCurrentFile : Boolean;
			key: Str255;
	begin
		{$unused reply}
		{$unused refcon}
		
		(* Get the parameters out of the event. Note that this code throws away
			errors left right and centre, mainly because we have a good default
			behaviour, which we'll execute if we get any errors.
		*)
		useFileSpec := true;
		useCurrentFile := false;
		if AEGetParamPtr(appleEvent, keyEditPreferenceDestination, typeFSS, junkType,
											@fileSpec, sizeof(fileSpec), junkActualSize) <> noErr then begin
			(* No file given, obviously can't use the file spec. Now let's try to get the
				parameter as a Boolean.
			*)
			useFileSpec := false;
			useCurrentFile := true;
			if AEGetParamPtr(appleEvent, keyEditPreferenceDestination, typeBoolean, junkType, 
											@fileSpec, sizeof(fileSpec), junkActualSize) = noErr then begin
				(* No Boolean in the parameter, so we shouldn't use the current file.
					The default Internet Preferences file will be used instead.
				*)
				useCurrentFile := false;
			end; (* if *)
		end; (* if *)
		if AEGetParamPtr(appleEvent, keyDirectObject, typeChar, junkType, 
										@key[1], 255, actualSize) <> noErr then begin
			actualSize := 0;
		end; (* if *)
		err := AEGotRequiredParams(appleEvent);
		
		(* Everything is cool, let's start doing the job. *)
		if err = noErr then begin
			key[0] := chr(actualSize);
			err := DoEditPreference(key, fileSpec, useFileSpec, useCurrentFile);
		end; (* if *)
		
		HandleAEEditPreferenceEvent := err;
	end; (* HandleAEEditPreferenceEvent *)
{$POP}

	(* **** Code to Actually Get a URL ***** *)
	
	function DoGetURL(urlText : Handle) : OSStatus;
		(* This routine calls IC to actually perform the GURL event action,
			ie do all the hard work finding and launching the configured
			application.  We use CreateSpecifiedInstance to
			make sure that we have a properly specified instance
			even if no document is open.
		*)
		var
			err : OSStatus;
			inst : ICInstance;
			startSelection : longint;
			endSelection : longint;
			s : SInt8;
	begin
		(* Make sure we have a specified instance first. *)
		err := CreateSpecifiedInstance(inst);

		(* Lock the urlText handle down and then call ICLaunchURL to do the hard work. *)
		if err = noErr then begin
			startSelection := 0;
			endSelection := GetHandleSize(urlText);
			s := HGetState(urlText);
			HLock(urlText);
  			err := ICLaunchURL(inst, '', urlText^, GetHandleSize(urlText),
  														startSelection, endSelection);
  			HSetState(urlText, s);
		
			DestroySpecifiedInstance(inst);
		end; (* if *)

		DoGetURL := err;
	end; (* DoGetURL *)

	(* **** The GURL Event Handler ***** *)

{$PUSH}
{$R-}
	function HandleAEGurlEvent(var appleEvent : AppleEvent; var reply: AppleEvent; refcon: longint): OSErr;
		var
			err: OSStatus;
			junk : OSStatus;
			urlDesc : AEDesc;
			errStr : Str255;
	begin
		{$unused reply}
		{$unused refcon}
		
		urlDesc.descriptorType := typeNull;
		urlDesc.dataHandle := nil;
		err := AEGetParamDesc(appleEvent, keyDirectObject, typeChar, urlDesc);
		if err = noErr then begin
			err := AEGotRequiredParams(appleEvent);
		end; (* if *)
		
		(* Everything is cool, let's start doing the job. *)
		if err = noErr then begin
			err := DoGetURL(urlDesc.dataHandle);
		end; (* if *)
		
		(* If the sender is expecting a reply, set it up for them. *)
		if reply.dataHandle <> nil then begin
			if err <> noErr then begin
				errStr := NewLookupError(rMainErrorList, err); 
				if errStr <> '' then begin
					junk := AEPutParamPtr(reply, keyErrorString, typeChar, @errStr[1], length(errStr));
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		(* Clean up. *)		
		junk := AEDisposeDesc(urlDesc);
		
		HandleAEGurlEvent := err;
	end; (* HandleAEGurlEvent *)
{$POP}

	procedure InstallMiscAppleEventHandlers;
		var
			junk: OSStatus;
	begin
		junk := AEInstallEventHandler(kICEditPreferenceEventClass, kICEditPreferenceEvent, 
										NewAEEventHandlerProc(HandleAEEditPreferenceEvent), 0, false);
		junk := AEInstallEventHandler(kGURLEventClass, kGURLEvent, 
										NewAEEventHandlerProc(HandleAEGurlEvent), 0, false);
	end; (* InstallMiscAppleEventHandlers *)

end. (* ICMiscAppleEvents *)
